
# X-13ARIMA-SEATS
The TimeSeriesEcon.X13 module provides a Julia interface to the X-13ARIMA-SEATS seasonal adjustment program. The relevant spec structure can be built in julia and the results are loaded into TSeries, MVTSeries, and table-like Workspace objects.

## General use
Running X13 involves three steps:

1. Creating a new spec object by either initializing it with a TSeries (i.e. `X13.newspec(ts)`) or with an [X13.series](#TimeSeriesEcon.X13.series-Union{Tuple{TSeries{F,%20T}%20where%20T<:Number},%20Tuple{F}}%20where%20F<:Frequency) object (i.e. `X13.newspec(X13.series(ts; span=1967Q1:1979Q4, title="My TSeries"))`).

2. Adding additional components to the spec with the specified bang functions (i.e. `X13.x11!(spec)` or `X13.regression!(spec, variables = :const)`). These can also be added in the first step by specifying them as keyword arguments, i.e. `spec = X13.newspec(ts; x11 = X13.x11(), transform=X13.transform(; func=:log))``

3. Running X13 with `X13.run(spec)`. This run returns an [X13result](#TimeSeriesEcon.X13.X13result) object.

Below is a simple example:

```julia 
# construct TSeries
seasonal = repeat([0.9, 1.0, 0.7, 1.5], 20)
trend = 53 .+ collect(101:180) .^ 1.05
ts = TSeries(1967Q1, trend .* seasonal)

# Construct spec and run X13
spec = X13.newspec(ts)
X13.transform!(spec, func=:log)
X13.regression!(spec, variables = :const)
X13.x11!(spec)
res = X13.run(spec)

# seasonally-adjusted series
println(X13.descriptions(res).series.d11)
println(res.series.d11[1967Q1:1967Q4])

# seasonal-adjustment factors
println(X13.descriptions(res).series.d10)
println(res.series.d10[1967Q1:1967Q4])

# OUTPUT:
# X11: final seasonally adjusted series
# 4-element TSeries{Quarterly} with range 1967Q1:1967Q4:
#   1967Q1 : 184.599195705163
#   1967Q2 : 186.032683808336
#   1967Q3 : 187.550857059321
#   1967Q4 : 188.83696094707

# X11: final seasonal factors
# 4-element TSeries{Quarterly} with range 1967Q1:1967Q4:
#   1967Q1 : 0.878623915125452
#   1967Q2 : 0.975837118555581
#   1967Q3 : 0.68249633096449
#   1967Q4 : 1.46304978198987
```

All of the available spec components have particular keyword arguments for specifying their options. These are documented extensively below.

### Convenience functions
There are two main convenience functions. The first is the `descriptions(res::X13result)` function which returns a structure similar to the results object, but holding string descriptions of the series/tables, rather than their contents.

The second is the `deseasonalize(ts)` and `deseasonalize!(ts)` functions. This function runs a simple specification with only the X11 component and returns the d11 series from the output. Note that this may not be the correct approach for some data. See [`deseasonalize!`](#TimeSeriesEcon.X13.deseasonalize!-Tuple{TSeries}) and the [X11 spec](#TimeSeriesEcon.X13.x11-Tuple{}) documentation for more details.

### Using your own binary
The TimeSeriesEcon package comes bundled with a binary for X-13ARIMA-SEATS but can also be called with a custom binary. To call a specific binary, the `:x13path` variable can be set with `TimeSeriesEcon.setoption(:x13path, "path/to/x13.exe")`. Note that many of the checks performed assume an that X13-ARIMA-SEATS is used, so earlier versions (i.e. X-12) may not be compatible with the spec file(s) produced. In addition, only the ascii-output version of X-13 is supported. The default path is `""`.

### X-13ARIMA-SEATS Manual
Official documentation for X13-ARIMA-SEATS can be found on the [US Census Bureau website](https://www.census.gov/data/software/x13as.References.html#par_list_410459519).

In particular, much of the documentation below makes further reference to the [X-13ARIMA-SEATS Reference Manual](https://www2.census.gov/software/x-13arima-seats/x-13-data/documentation/docx13as.pdf)


## Function / Struct Reference
```@autodocs
Modules = [X13]
Order   = [:function, :type]
```